
(*: Name : SimplePolygonTriangulation` *)

(*: Title : SimplePolygonTriangulation *)

(*: Author : Martin Kraus *)

(*: Package Version : 1.2 *)

(*: Mathematica Version : 4.0 *)

(*: Summary : 
      "This package provides functions to triangulate simple \
polygons,i.e.polygons without intersecting edges or holes, in two and three \
dimensions." *)

(*: History : 
      "Created August 1999 by Martin Kraus (mk, \
mailto:Martin_Kraus_Germany@yahoo.com), 
09/06/99 mk version 1.1: cleaned up examples and improved dropping of last \
point if same as first
07/01/99 mk version 1.2: new context PolygonTriangulation`" *)

(*: Keywords : "geometry,polygon,triangulation,non-convex,simple" *)

(*: Source : 
      "'Mathematica Graphics: Techniques and Applications' by Tom \
Wickham-Jones,TELOS/Springer-Verlag 1994 (especially the package \
NonConvexTriangulate.m); 'Computational Geometry in C' by Joseph \
O'Rourke,Cambridge Univeristy Press,1994" *)

(*: Warnings : See Limitation. *)

(*: Limitation : 
      "This package will not work with self-intersecting polygon or non-flat \
polygons in three dimensions. (Polygons in three dimensions are projected \
onto a plane. The functions will work correctly iff the projected polygon is \
not self-intersecting." *)

(*: Discussion :
      "This implementation uses a trapezoidalization in order to partition a \
simple polygon into monotone polygons (see option GraphicalDebugging). The \
line-sweep algorithm is implemented with O(N^2) (N:number of vertices) \
instead of O(N log N) as lists are used for several data structures (monotone \
polygons,edges intersected by the sweep line) instead of balanced trees. \
After the trapezoidalization the monotone polygons are triangulated again \
with O(N^2) for the same reason with respect to the resulting triangles. \
Polygons in three-dimensions are projected onto a plane before the \
triangulation.The plane is determined either by a specified normal (e.g. \
second argument of Polygon primitives) or by the cross product of the first \
two edges of the polygon." *)

BeginPackage["PolygonTriangulation`SimplePolygonTriangulation`"];

(* Usage Messages *)

SimplePolygonTriangulation::usage = 
    "SimplePolygonTriangulation[pts] returns a triangulation of the list of \
two-dimensional points pts that maintains the outline. The triangulation \
consists of a list of lists of three indices. Each triple of indices \
describes one triangle.";

SimplePolygonTriangulation3D::usage = 
    "SimplePolygonTriangulation3D[pts] returns a triangulation of the list of \
three-dimensional points pts that maintains the outline. The triangulation \
consists of a list of lists of three indices. Each triple of indices \
describes one triangle.";

TriangulateSimplePolygons::usage = 
    "TriangulateSimplePolygons[expr] returns expr with triangulated polygons \
assuming all polygons in expr are two-dimensional and simple.";

TriangulateSimplePolygons3D::usage = 
    "TriangulateSimplePolygons3D[expr] returns expr with triangulated \
polygons assuming all polygons in expr are three-dimensional and simple (see \
also option SimplePolygons).";

GraphicalDebugging::usage = 
    "GraphicalDebugging is an option for SimplePolygonTriangulation and \
related functions. GraphicalDebugging->True will show a sequence of Graphics \
showing the progress of the triangulation.";

PlaneNormal::usage = 
    "PlaneNormal is an option for SimplePolygonTriangulation3D specifying the \
normal vector of the plane used to calculate the two-dimensional projection \
of the three-dimensional points before the triangulation is done within the \
projection plane. PlaneNormal->Automatic will automatically determine a \
normal vector.";

TriangleEdges::usage = 
    "TriangleEdges is an option for TriangulateSimplePolygons and \
TriangulateSimplePolygons3D. TriangleEdges->Automatic will suppress the \
creation of any new edges during the triangulation. TriangleEdges->All will \
return triangles without explicit edge specification in the third argument of \
Polygon primitives.";

QuadrangleQuartering::usage = 
    "QuadrangleQuartering is an option for TriangulateSimplePolygons3D. \
QuadrangleQuartering->False will triangle quadrangles like all other \
polygons. QuadrangleQuartering->True will quater each quadrangle into four \
triangles each containing the center of mass of the quadrangle.";

SimplePolygons::usage = 
    "SimplePolygons is an option for TriangulateSimplePolygons3D. \
SimplePolygons->All will triangulate all polygons. SimplePolygons->Automatic \
will only triangulate Polygon primitives with a second argument specifying \
the normal vector of the polygon (this second argument may be Automatic).";

(* Error Messages *)

SimplePolygonTriangulation::internal = "An internal error (code `1`) occured.
	Please check the polygon for self-intersections.";

SimplePolygonTriangulation::winding = 
    "SimplePolygonTriangulation can only triangulate polygons with winding \
number 1 or -1.";

Begin["`Private`"];

Off[General::spell1];

(* Options *)

Options[SimplePolygonTriangulation] = {GraphicalDebugging -> False};

Options[SimplePolygonTriangulation3D] = {PlaneNormal -> Automatic};

Options[TriangulateSimplePolygons] = {};

Options[TriangulateSimplePolygons3D] = {GraphicalDebugging -> False, 
      SimplePolygons -> All, TriangleEdges -> Automatic, 
      QuadrangleQuartering -> False};

(* Auiliary Functions *)

(* "order: two-dimensional points,vertices,edges,monotone polygons \
(monos),triangles,simple polygons,graphics" *)

(* "the normalized vector or {0, 0, 0}" *)

normalize[vec_] := 
    Module[{length}, length = N[Sqrt[vec.vec]]; 
      If[length <= $MachineEpsilon, {0, 0, 0}, vec/length]];

(* "determines a plane approximating a set of points, the plane is given as a \
list of one point, the normal vector, and two basis vectors for a plane (the \
first basis vector together with the second basis vector and the normal \
vector form an orthonormal right-hand system)" *)

getPlane[points_] := 
    Module[{normal, numberOfPoints, pointsLeft, pointsLeftLeft, crossProducts,
         basis1, basis2}, numberOfPoints = Length[points]; 
      pointsLeft = RotateLeft[points]; 
      pointsLeftLeft = RotateLeft[pointsLeft]; 
      crossProducts = 
        Table[normalize[
            Cross[pointsLeft[[i]] - points[[i]], 
              pointsLeftLeft[[i]] - pointsLeft[[i]]]], {i, 1, 
            numberOfPoints}]; normal = normalize[Plus @@ crossProducts]; 
      While[normal == {0, 0, 0} && Length[crossProducts] > 0, 
        normal = First[crossProducts]; 
        crossProducts = Drop[crossProducts, 1]]; 
      If[normal == {0, 0, 0}, 
        Return[{Plus @@ N[points]/numberOfPoints, {0, 0, 0}, {0, 0, 0}, {0, 0,
               0}}]]; basis1 = normalize[First[NullSpace[{N[normal]}]]]; 
      basis2 = normalize[Cross[normal, basis1]]; {Plus @@ N[points]/
          numberOfPoints, normal, basis1, basis2}];

(* "projects a three-dimensional point to a plane which is given by a list of \
one point, a normal vector, and two basis vectors (see getPlane)" *) 
projectPointToPlane[point_, plane_] := 
    Module[{planePoint, basis1, basis2}, planePoint = plane[[1]]; 
      basis1 = plane[[3]]; 
      basis2 = plane[[4]]; {(point - planePoint).basis1, (point - 
              planePoint).basis2}];

(* "returns the three-dimensional point of a point defined on a plane (see \
getPlane for the plane argument)" *)

embedPointIn3D[point_, plane_] := 
    Module[{planePoint, basis1, basis2}, planePoint = plane[[1]]; 
      basis1 = plane[[3]]; basis2 = plane[[4]]; 
      point[[1]]*basis1 + point[[2]]*basis2 + planePoint];

(* "the angle at the two-dimensional point p2, positive for anti-clockwise \
turn" *)
angle[p1_, p2_, p3_] := 
    Module[{dp1, dp2, factor, result}, dp1 = p2 - p1; dp2 = p3 - p2; 
      factor = N[Sqrt[(dp1.dp1)*(dp2.dp2)]]; 
      If[factor <= $MachineEpsilon, Return[0]]; 
      result = Re[ArcSin[(dp1[[1]]*dp2[[2]] - dp1[[2]]*dp2[[1]])/factor]]; 
      If[dp1.dp2 >= 0, result, N[Mod[2.  Pi - result, 2. Pi] - Pi]]];

(* "the windingNumber of a list of two-dimensional points. Positive for \
anti-clockwise turns." *)

windingNumber[points_] := 
    Module[{rotatedLeft, rotatedLeftLeft, numberOfPoints}, 
      numberOfPoints = Length[points]; rotatedLeft = RotateLeft[points]; 
      rotatedLeftLeft = RotateLeft[rotatedLeft]; 
      Round[Sum[
            angle[points[[i]], rotatedLeft[[i]], rotatedLeftLeft[[i]]], {i, 1,
               numberOfPoints}]/ (2. Pi)]];

(* "calculates the (positive) index distance between two vertex indices" *)
  vertexIndexDistance[a_, b_] := 
    Module[{amod, bmod}, amod = Mod[a - 1, numberOfVertices]; 
      bmod = Mod[b - 1, numberOfVertices]; 
      Min[Abs[amod - bmod], numberOfVertices - Abs[amod - bmod]]];

(* "calculates the "smaller" of two neighboring vertex indices" *) 
vertexSmallerIndex[a_, b_] := 
    Module[{}, 
      If[(a == numberOfVertices && b == 1) || (a == 1 && 
              b == numberOfVertices), numberOfVertices, Min[a, b]]];

(* "navigation functions operating on indices" *)

vertexNext[a_] := Mod[(a + 1) - 1, numberOfVertices] + 1;
vertexPrevious[a_] := Mod[(a - 1) - 1, numberOfVertices] + 1;

(* "vertexLeftQ[a,b,c] returns True if vertex with index c is to the left of \
the line through a and b" *) 
vertexLeftQ[a_, b_, 
      c_] := ((vertices[[b, 1]] - vertices[[a, 1]])*(vertices[[c, 2]] - 
                vertices[[b, 2]]) - (vertices[[b, 2]] - 
                vertices[[a, 2]])*(vertices[[c, 1]] - vertices[[b, 1]]) > 0);

(* "vertexAboveQ[a,b] returns True if the vertex with index a is above the \
vertex with index b (greater y-coordinate) or if the y coordinates are equal \
and the x-coordinate of a is greater or equal to the x-coordinate of b. Thus, \
two different points are never on the same height (effect of an infinitesimal \
anti-clockwise rotation).vertexAboveQ should always be used in order to \
decide which point is higher in order to guarantee a consistent order; in \
this case,vertexAboveQ should not be called with identical arguments." *)

vertexAboveQ[a_, 
      b_] := (Order[mirroredVertices[[a]], mirroredVertices[[b]]] <= 0);

(* "determine the cusp type. The discrimination between interior and exterior \
cusps is done implicitly as the edges are always separating interior and \
exterior regions (no overlapping edges are allowed). This function returns \
None if a vertex is no cusp, Downward if it is a downward cusp and Upward if \
it is a upward cusp." *)

vertexCuspType[a_] := 
    If[vertexAboveQ[vertexPrevious[a], a], 
      If[vertexAboveQ[a, vertexNext[a]], None, Downward], 
      If[vertexAboveQ[vertexNext[a], a], None, Upward]];

(* "processes one vertex" *)

vertexProcess[v_] := 
    Module[{pos, type, previous, next, below, above, right, left, monoPos, 
        monoList}, pos = edgePosition[v]; type = vertexCuspType[v]; 
      next = vertexNext[v]; previous = vertexPrevious[v]; 
      If[vertexAboveQ[previous, next], above = previous; below = next, 
        below = previous; 
        above = next];(*Print[{v, pos, type, previous, next, below, above}];*)(* 
        update edges and montone polygons *)
      monoPos = Ceiling[pos/2];
      If[type === None, 
        If[edges[[pos, 1]] != v && edges[[pos, 2]] != v, 
          Throw["msgpro1", SimplePolygonTriangulation]]; 
        monoList = openMonotonePolygons[[monoPos]];(* 
          if there are two monotone polygons we close one *)
        
        If[Length[monoList] > 1, 
          If[OddQ[pos],(* left border *) monoClose[monoList[[1]], v]; 
            monoList = {monoList[[2]]}, (* right border *) 
            monoClose[monoList[[2]], v]; monoList = {monoList[[1]]}]];(* 
          now monoList is of length 1 and we can add the vertex to the \
(remaining) monotone polygon *) 
        If[OddQ[pos],(* left border *) 
          monoList = {monoAppendLeft[monoList[[1]], v]},(* right border *) 
          monoList = {monoAppendRight[monoList[[1]], v]}]; 
        openMonotonePolygons[[monoPos]] = monoList; 
        edgeReplace[v, below, pos], 
        If[type === Upward, 
          If[OddQ[pos],(* exterior cusp : we insert a new monotonePolygon *)
              openMonotonePolygons = 
              Insert[openMonotonePolygons, {monoCreate[v]}, monoPos],(* 
              interior cusp : 
                  either separate two monotone polygons or split one *) 
            monoList = openMonotonePolygons[[monoPos]]; 
            If[Length[monoList] > 1, 
              monoList = {monoAppendRight[monoList[[1]], v], 
                  monoAppendLeft[monoList[[2]], v]}; 
              openMonotonePolygons[[monoPos]] = {monoList[[1]]}; 
              openMonotonePolygons = 
                Insert[openMonotonePolygons, {monoList[[2]]}, monoPos + 1],
              (* 
                length of monoList is 1 : 
                  we have to connect the lower reaching chain to our cusp and \
create a new monotone polygon *) 
              If[vertexAboveQ[Last[monoList[[1, 2]]], 
                  Last[monoList[[1, 1]]]],(* left chain is lower *) 
                openMonotonePolygons[[monoPos]] = {monoAppendLeft[
                      monoList[[1]], v]}; 
                openMonotonePolygons = 
                  Insert[openMonotonePolygons, {monoAppendRight[
                        monoCreate[Last[monoList[[1, 1]]]], v]}, monoPos],
                (* right chain is lower *) 
                openMonotonePolygons[[monoPos]] = {monoAppendRight[
                      monoList[[1]], v]}; 
                openMonotonePolygons = 
                  Insert[openMonotonePolygons, {monoAppendLeft[
                        monoCreate[Last[monoList[[1, 2]]]], v]}, 
                    monoPos + 1]]]]; 
          If[vertexLeftQ[previous, v, next], left = next; right = previous, 
            left = previous; right = next]; edgeInsert[v, right, pos]; 
          edgeInsert[v, left, pos], 
          If[type =!= 
                Downward || (edges[[pos, 1]] != v && 
                  edges[[pos, 2]] != v) || (edges[[pos + 1, 1]] != v && 
                  edges[[pos + 1, 2]] != v), 
            Throw["msgpro2", SimplePolygonTriangulation]]; 
          If[OddQ[pos],(* exterior cusp : close all monotone polygons *)
              monoList = openMonotonePolygons[[monoPos]]; 
            Map[(monoClose[#, v]) &, monoList]; 
            openMonotonePolygons = Drop[openMonotonePolygons, {monoPos}],(* 
              interior cusp : 
                  first we have to close monotone polygons on the left and/
                      or right. Then we join the remaining two in a list *) 
            monoListLeft = openMonotonePolygons[[monoPos]]; 
            monoListRight = openMonotonePolygons[[monoPos + 1]]; 
            If[Length[monoListLeft] > 1, monoClose[monoListLeft[[-1]], v]]; 
            If[Length[monoListRight] > 1, monoClose[monoListRight[[1]], v]]; 
            openMonotonePolygons[[monoPos]] = {monoAppendRight[
                  monoListLeft[[1]], v], 
                monoAppendLeft[monoListRight[[-1]], v]}; 
            openMonotonePolygons = 
              Drop[openMonotonePolygons, {monoPos + 1}]]; edgeDelete[pos]; 
          edgeDelete[pos]]]];

(* "returns the x-coordinate of the intersection of a line specified by a and \
b and a horizontal line at y. If there is no intersection, False is returned. \
If the lines are identical, True is returned." *)

xCoordinateOfIntersection[y_, a_, b_] := 
    Module[{x1, y1, x2, y2}, {x1, y1} = vertices[[a]]; {x2, y2} = 
        vertices[[b]]; 
      If[y2 == y1, 
        If[y1 == y, True, False], ((x2 - x1)*(y - y1))/(y2 - y1) + x1]];

(* "True if vertex b is reflex (internal angle >= \[Pi]) with respect to a \
and c, employing the convention that the interior of the polygon is to the \
left when going from a to b" *) 
vertexReflexQ[a_Integer, b_Integer, c_Integer] := 
    Module[{x1, y1, x2, y2}, {x1, y1} = 
        vertices[[b]] - vertices[[a]]; {x2, y2} = 
        vertices[[c]] - vertices[[b]]; (x1 y2 - x2 y1) <= 0];

(* "inserts an edge between a and b into the list of edges at position at" *)

edgeInsert[a_, b_, at_] := 
    Module[{edge, v1, v2, x1, x2}, 
      If[at < 1 || at > numberOfEdges + 1, 
        Throw["msgins1", SimplePolygonTriangulation], 
        If[b == vertexNext[a], v1 = a; v2 = b, 
          If[a == vertexNext[b], v1 = b; v2 = a, 
            Throw["msgins", SimplePolygonTriangulation], v1 = a; 
            v2 = vertexNext[a]]]; x1 = vertices[[v1, 1]]; 
        x2 = vertices[[v2, 1]]; edge = {v1, v2, Min[x1, x2], Max[x1, x2]}; 
        edges = Insert[edges, edge, at]; 
        numberOfEdges = numberOfEdges + 1;]];

(* "deletes an edge at position at" *)

edgeDelete[at_] := 
    Module[{}, 
      If[at < 1 || at > numberOfEdges, 
          Throw["msgdel", SimplePolygonTriangulation], 
          edges = Drop[edges, {at}]; numberOfEdges = numberOfEdges - 1];];

(* "replaces an edge at position at by the new edge from a to b" *)

edgeReplace[a_, b_, at_] := Module[{}, edgeDelete[at]; edgeInsert[a, b, at]];

(* "returns the index of the left-most edge containing vertex v or the index \
of the edge next to it at the right-hand side." *)

edgePosition[v_] := 
    Module[{left, right, middle, xMiddle, result, xMin, xMax, x, y}, 
      If[numberOfEdges == 0, Return[1]]; {x, y} = vertices[[v]];(* 
        binary search : 
            interval is from left (inclusive) to right (exclusive) *)
      
      left = 1; right = numberOfEdges + 1; 
      While[right - left > 0, middle = Floor[(right + left)/2]; 
        xMin = edges[[middle, 3]]; xMax = edges[[middle, 4]]; 
        If[xMax < x, left = middle + 1, 
          If[xMin > x, 
            right = middle,(* test, whether v is a vertex of edge middle *)
  
                      
            If[edges[[middle, 1]] == v || edges[[middle, 2]] == v,(* it is, 
                but edge middle - 1 might also contain v! *) 
              If[middle > 
                    left && (edges[[middle - 1, 1]] == v || 
                      edges[[middle - 1, 2]] == v), left = middle - 1; 
                right = middle - 1, left = middle; right = middle],(* OK, 
                we know that xMin <= x <= 
                  xMax and v is not a vertex of edge middle *)
              
              xMiddle = 
                xCoordinateOfIntersection[y, edges[[middle, 1]], 
                  edges[[middle, 2]]]; 
              If[xMiddle === True || xMiddle === False,(* 
                  we assume that this edge is on the sweep line; therefore, 
                  v is part of the edge but not a vertex, 
                  i.e. there is a self - intersection, which is forbidden *)
                  Throw["msgpos1", SimplePolygonTriangulation]; left = middle;
                 right = middle, 
                If[xMiddle <= x, left = middle + 1, right = middle]]]]]]; 
      If[left != right, Throw["msgpos2", SimplePolygonTriangulation]]; 
      If[! IntegerQ[right], Throw["msgpos3", SimplePolygonTriangulation]; 
        right = 1]; Return[right]];

(* "creates a new monotonePolygon" *)

monoCreate[v_] := Module[{}, monotonePolygon[{v}, {v}]];


(* "appending vertices to chains of monotone polygons" *)

monoAppendLeft[monotonePolygon[left_, right_], v_] := 
    monotonePolygon[Append[left, v], right];

monoAppendRight[monotonePolygon[left_, right_], v_] := 
    monotonePolygon[left, Append[right, v]];

(* "adds v to the end of both chains and appends the monotonePolygon to the \
closedMonotonePolygons" *)
  monoClose[monotonePolygon[left_, right_], v_] := 
    Module[{}, 
      AppendTo[closedMonotonePolygons, 
          monotonePolygon[Append[left, v], Append[right, v]]];];

(* "returns {nextV, nextVLeftQ,newLeftPosition, newRightPosition} with nextV \
being the vertex next below the vertices at positions leftPosition in the \
left chain and at rightPosition in the right chain. If there is no vertex \
left nextV is set to 0. Note: The last elements of the chains are \
identical,we have to take care that this element is not returned twice." *) 
monoNextV[monotonePolygon[leftChain_, rightChain_], leftPosition_, 
      rightPosition_] := 
    Module[{}, 
      If[leftPosition >= Length[leftChain], 
        If[rightPosition + 1 >= Length[rightChain], {0, False, leftPosition, 
            rightPosition}, {rightChain[[rightPosition + 1]], False, 
            leftPosition, rightPosition + 1}], 
        If[rightPosition >= Length[rightChain], 
          If[leftPosition + 1 >= Length[leftChain], {0, False, leftPosition, 
              rightPosition}, {leftChain[[leftPosition + 1]], True, 
              leftPosition + 1, rightPosition}], 
          If[vertexAboveQ[leftChain[[leftPosition + 1]], 
              rightChain[[rightPosition + 1]]], {leftChain[[leftPosition + 
                    1]], True, leftPosition + 1, 
              rightPosition}, {rightChain[[rightPosition + 1]], False, 
              leftPosition, rightPosition + 1}]]]];

(* "triangulates a monotone Polygon and appends the resulting triangles to \
the list of triangles." *)

monoTriangulate[mono : monotonePolygon[leftChain_, rightChain_]] := 
    Module[{reflexChain, reflexChainLeftQ, v, vLeftQ, leftPos, rightPos, 
        flagAdvanceV}, 
      If[vertexAboveQ[leftChain[[2]], rightChain[[2]]], 
        reflexChain = Take[leftChain, 2]; reflexChainLeftQ = True; 
        If[vertexAboveQ[leftChain[[3]], rightChain[[2]]], v = leftChain[[3]]; 
          vLeftQ = True; leftPos = 3; rightPos = 1, v = rightChain[[2]]; 
          vLeftQ = False; leftPos = 2; rightPos = 2], 
        reflexChain = Take[rightChain, 2]; reflexChainLeftQ = False; 
        If[vertexAboveQ[rightChain[[3]], leftChain[[2]]], v = rightChain[[3]];
           vLeftQ = False; rightPos = 3; leftPos = 1, v = leftChain[[2]]; 
          vLeftQ = True; rightPos = 2; leftPos = 2]]; flagAdvanceV = False; 
      While[If[graphicalDebugging, 
          graphicsShow[0, 
            monoGraphics[mono, v, reflexChain, leftPos, rightPos]]]; v != 0, 
        If[Xor[reflexChainLeftQ, vLeftQ],(* 
            v and reflexChain on opposite chains : 
                create triangle between v and first two points of reflex \
chain *) If[vLeftQ, triangleAppend[{v, reflexChain[[2]], reflexChain[[1]]}], 
            triangleAppend[{v, reflexChain[[1]], reflexChain[[2]]}]]; 
          reflexChain = Rest[reflexChain]; 
          If[Length[reflexChain] == 1, flagAdvanceV = True],(* 
            v and reflexChain are on the same chains *) 
          If[vLeftQ,(* v and reflex chain on left chain *) 
            If[! vertexReflexQ[reflexChain[[-2]], reflexChain[[-1]], v],(* 
                vertex above v is strictly convex : we make a triangle *) 
              triangleAppend[{reflexChain[[-2]], reflexChain[[-1]], v}]; 
              reflexChain = Drop[reflexChain, -1]; 
              If[Length[reflexChain] == 1, flagAdvanceV = True],(* 
                vertex above v is reflex : we append v to the reflex chain *) 
              AppendTo[reflexChain, v]; flagAdvanceV = True],(* 
              v and reflex chain on right chain *) 
            If[! vertexReflexQ[v, reflexChain[[-1]], reflexChain[[-2]]],(* 
                vertex above v is strictly convex : we make a triangle *) 
              triangleAppend[{v, reflexChain[[-1]], reflexChain[[-2]]}]; 
              reflexChain = Drop[reflexChain, -1]; 
              If[Length[reflexChain] == 1, flagAdvanceV = True],(* 
                vertex above v is reflex : we append v to the reflex chain *) 
              AppendTo[reflexChain, v]; flagAdvanceV = True]]]; (* 
          advance v if needed *) 
        If[flagAdvanceV, 
          If[Length[reflexChain] == 1, (* 
              reflexChain is empty : setup new reflex chain, 
              just like at the start (note : 
                    the first element of the relfex chain does not have to be \
on the same chain as the rest of the reflex chain!) *) 
            reflexChain = {reflexChain[[1]], v}; 
            reflexChainLeftQ = vLeftQ]; {v, vLeftQ, leftPos, rightPos} = 
            monoNextV[mono, leftPos, rightPos]; flagAdvanceV = False]]; 
      If[Length[reflexChain] > 2, 
        Throw["msgmot1", SimplePolygonTriangulation]]];

(* "creates graphics for the monotone polygons" *)

monoGraphics[monotonePolygon[leftChain_, rightChain_], v_, reflexChain_, 
      leftPos_, rightPos_] := 
    Module[{primitivesPositions, primitivesReflexChain, primitivesVertex}, 
      primitivesPositions = {Thickness[0.02], RGBColor[1, 1, 0], 
          PointSize[0.05], Point[vertices[[leftChain[[leftPos]]]]], 
          Point[vertices[[rightChain[[rightPos]]]]]}; 
      primitivesVertex = 
        If[v != 0, {Thickness[0.02], RGBColor[1, 0, 0], PointSize[0.03], 
            Point[vertices[[v]]]}, {}]; 
      primitivesReflexChain = {Dashing[{0.01, 0.05}], RGBColor[0, 0.8, 0], 
          Thickness[0.02], Line[Part[vertices, reflexChain]]}; 
      Graphics[{primitivesReflexChain, primitivesPositions, 
          primitivesVertex}]];

(* "append a new triangle to the list of triangles, if it is non-vanishing." *) 
triangleAppend[{a_Integer, b_Integer, c_Integer}] := 
    Module[{}, If[! vertexReflexQ[a, b, c], AppendTo[triangles, {a, b, c}]]];

(* "triangulates one polygon in three dimensions using the options of \
TriangulatesSimplePolygons3D.Attention:This function needs the value of \
numberOfVertices even after SimplePolygonTriangulation." *)
  triangulateSimplePolygon3D[
      poly : Polygon[points_:{}, normal_:None, edges_:{}], opts___] := 
    Module[{graphicalDebugging, planeNormal, triangleEdges, 
        quadrangleQuartering, fullEdges, realNormal, center, trianglePoints, 
        triangulation, triangulationEdges, simplePolygons, 
        i}, {graphicalDebugging, triangleEdges, quadrangleQuartering, 
          simplePolygons} = {GraphicalDebugging, TriangleEdges, 
              QuadrangleQuartering, SimplePolygons} /. {opts} /. 
          Options[TriangulateSimplePolygons3D];
      If[Length[points] <= 
            3 || (normal === None && simplePolygons === Automatic), 
        Return[poly]];
      If[normal === None, realNormal = Automatic, realNormal = normal];
      If[Length[points] != Length[edges], 
        fullEdges = Table[True, {Length[points]}], fullEdges = edges];
      If[Length[points] == 4 && quadrangleQuartering === True, 
        center = Plus @@ points/4.;
        trianglePoints = {{points[[1]], points[[2]], center}, {points[[2]], 
              points[[3]], center}, {points[[3]], points[[4]], 
              center}, {points[[4]], points[[1]], center}};
        Return[
          Table[If[triangleEdges === All && normal === None, 
              Polygon[trianglePoints[[i]]], 
              If[triangleEdges === All, 
                Polygon[trianglePoints[[i]], realNormal], 
                Polygon[trianglePoints[[i]], 
                  realNormal, {fullEdges[[i]], False, False}]]], {i, 1, 
              4}]]];
      triangulation = 
        SimplePolygonTriangulation3D[points, 
          GraphicalDebugging -> graphicalDebugging, 
          PlaneNormal -> realNormal];
      If[triangulation == {}, Return[poly]];
      triangulationEdges = 
        Map[({If[vertexIndexDistance[#[[1]], #[[2]]] != 1, False, 
                  fullEdges[[vertexSmallerIndex[#[[1]], #[[2]]]]]], 
                If[vertexIndexDistance[#[[2]], #[[3]]] != 1, False, 
                  fullEdges[[vertexSmallerIndex[#[[2]], #[[3]]]]]], 
                If[vertexIndexDistance[#[[3]], #[[1]]] != 1, False, 
                  fullEdges[[vertexSmallerIndex[#[[3]], #[[1]]]]]]}) &, 
          triangulation]; 
      trianglePoints = Map[Part[points, #] &, triangulation];
      Return[
        Table[If[triangleEdges === All && normal === None, 
            Polygon[trianglePoints[[i]]], 
            If[triangleEdges === All, 
              Polygon[trianglePoints[[i]], realNormal], 
              Polygon[trianglePoints[[i]], realNormal, 
                triangulationEdges[[i]]]]], {i, 1, 
            Length[trianglePoints]}]]];

graphicsShow[v_, moreGraphics_:Graphics[{}]] := 
    Module[{graphicsPolygon, graphicsVertex, graphicsEdges, 
        graphicsClosedMonotonePolygons, graphicsOpenMonotonePolygons, 
        graphicsTriangles, xMin, xMax, yMin, yMax}, 
      xMin = Min[Transpose[vertices][[1]]]; 
      xMax = Max[Transpose[vertices][[1]]]; 
      yMin = Min[Transpose[vertices][[2]]]; 
      yMax = Max[Transpose[vertices][[2]]]; 
      graphicsPolygon = 
        Graphics[{GrayLevel[0.5], Polygon[Append[vertices, First[vertices]]]},
           AspectRatio -> Automatic, 
          PlotRange -> {{xMin - 0.05(xMax - xMin), 
                xMax + 0.05(xMax - xMin)}, {yMin - 0.05(yMax - yMin), 
                yMax + 0.05(yMax - yMin)}}, Axes -> True]; 
      graphicsVertex = 
        If[v != 0, 
          Graphics[{Thickness[0.02], RGBColor[1, 0, 0], 
              Line[{{xMin, vertices[[v, 2]]}, {xMax, vertices[[v, 2]]}}], 
              PointSize[0.05], Point[vertices[[v]]]}], Graphics[{}]]; 
      graphicsClosedMonotonePolygons = 
        Graphics[{RGBColor[0, 0, 1], Thickness[0.02], 
            Map[{Line[Part[vertices, #[[1]]]], 
                  Line[Part[vertices, #[[2]]]]} &, closedMonotonePolygons]}]; 
      graphicsTriangles = 
        Graphics[{Thickness[0.02], 
            Map[Line[Append[Part[vertices, #], vertices[[#[[1]]]]]] &, 
              triangles]}]; 
      graphicsOpenMonotonePolygons = 
        Graphics[{Dashing[{0.02, 0.05}], RGBColor[1, 0, 1], Thickness[0.02], 
            Map[{Line[Part[vertices, #[[1]]]], 
                  Line[Part[vertices, #[[2]]]]} &, 
              Flatten[openMonotonePolygons]]}]; 
      graphicsEdges = 
        Graphics[{Dashing[{0.01, 0.05}], RGBColor[0, 0.8, 0], Thickness[0.02],
             Map[Line[{vertices[[#[[1]]]], vertices[[#[[2]]]]}] &, edges]}]; 
      Show[graphicsPolygon, graphicsVertex, graphicsClosedMonotonePolygons, 
        graphicsOpenMonotonePolygons, graphicsEdges, graphicsTriangles, 
        moreGraphics];(*Print[edges]*)];

(* Local (Static) Variables *)

(* "edges is a list of vertex indices, which describe the edges crossing the \
sweep line.Each element has this structure:{index of first vertex,index of \
second vertex (which is the vertexNext),minimum x coordinate of this \
edge,maximum x coordinate of this edge}" *)edges = {};

(* "numberOfEdges is equal to Length[edges]" *)
numberOfEdges = 0;

(* "vertices is a list of two-dimensional points \
({x-coordinate,y-coordinate})" *)
vertices = {};

(* "mirroredVertices is the same as vertices but with y-and x-coordinates \
interchanged." *)
mirroredVertices = {};

(* "openMonotonePolygons is a list of lists of monotonePolygons at positions \
corresponding to edges:the i-th list of monotone polygons is between the \
(2i-1)-th and (2i)-th edge.There are one or two monotone polygons per list.A \
monotonePolygon data structure has the form:monotonePolygon[{indices of \
vertices of left chain},{indices of vertices of right chain}].The first \
indices of the chains are identical for open monotone polygons." *)

openMonotonePolygons = {};

(* "closedMonotonePolygons is a list of closed monotonePolygons.A \
monotonePolygon data structure has the form:monotonePolygon[{indices of \
vertices of left chain},{indices of vertices of right chain}].The first and \
last indices of the chains are identical for closed monotone polygons." *)

closedMonotonePolygons = {};

(* "triangles is a list of lists of vertex indices,each element describing a \
triangle.The final value of triangles is the result of the triangulation." *)

triangles = {};

(* "graphicalDebugging is True if graphics should be produced" *)

graphicalDebugging = False;

(* Public Functions (see usage messages for comments) *)

SimplePolygonTriangulation[pts_, opts___] := 
    Module[{sortedVertexIndices, sweepIndex, a, b, c, x, y, winding, 
        catchResult}, {graphicalDebugging} = {GraphicalDebugging} /. {opts} \
/. Options[SimplePolygonTriangulation]; edges = {}; numberOfEdges = 0; 
      openMonotonePolygons = {}; closedMonotonePolygons = {}; 
      triangles = {};(* remove last point if duplicated *) 
      If[Chop[First[pts] - Last[pts]] == {0, 0}, vertices = Drop[pts, -1], 
        vertices = pts]; numberOfVertices = Length[vertices];
      winding = windingNumber[vertices];
      If[winding != -1 && winding != 1, 
        Message[SimplePolygonTriangulation::winding]; Return[{}]];
      If[numberOfVertices < 3, Return[{}]];
      If[numberOfVertices == 3, Return[{{1, 2, 3}}]];(* 
        mirror at the diagonal from down, left to up, right; 
        needed for vertexAboveQ *)
      
      mirroredVertices = 
        Transpose[{Transpose[vertices][[2]], Transpose[vertices][[1]]}]; 
      sortedVertexIndices = Sort[Range[Length[vertices]], vertexAboveQ]; 
      catchResult = 
        Catch[sweepIndex = 1; 
          While[sweepIndex <= numberOfVertices, 
            v = sortedVertexIndices[[sweepIndex]]; vertexProcess[v]; 
            If[! EvenQ[numberOfEdges], 
              Throw["msgswe1", SimplePolygonTriangulation]]; 
            If[graphicalDebugging, graphicsShow[v]]; 
            sweepIndex = sweepIndex + 1]; 
          Map[monoTriangulate, closedMonotonePolygons]; 
          If[winding == -1, triangles = Map[Reverse, triangles]]; True, 
          SimplePolygonTriangulation]; 
      If[True === catchResult, triangles, 
        Message[SimplePolygonTriangulation::internal, catchResult]; {}]];

SimplePolygonTriangulation3D[points3d_, opts___] := 
    Module[{points2d, plane, planeNormal, normal, basis1, 
        basis2}, {planeNormal} = {PlaneNormal} /. {opts} /. 
          Options[SimplePolygonTriangulation3D]; 
      If[MatchQ[planeNormal, {_, _, _}], normal = normalize[N[planeNormal]]; 
        basis1 = normalize[First[NullSpace[{normal}]]]; 
        basis2 = normalize[Cross[normal, basis1]]; 
        plane = {Plus @@ N[points3d]/Length[points3d], normal, basis1, 
            basis2}, plane = getPlane[points3d]]; 
      points2d = Map[projectPointToPlane[#, plane] &, points3d]; 
      SimplePolygonTriangulation[points2d, opts]];

TriangulateSimplePolygons[expr_, opts___] := 
    Module[{pts, triangulation}, 
      expr /. Polygon[
            pts_List] :> (triangulation = 
              SimplePolygonTriangulation[pts, opts]; 
            If[triangulation == {}, Polygon[pts], 
              Map[Polygon[Part[pts, #]] &, triangulation]])];

TriangulateSimplePolygons3D[expr_, opts___] := 
    Module[{poly}, 
      expr /. poly_Polygon :> triangulateSimplePolygon3D[poly, opts]];

On[General::spell1];

End[];

EndPackage[];

(*: Examples :*)

(*
<< SimplePolygonTriangulation`
*)
(*
(* simple example with SimplePolygonTriangulation : *)
*)
(*
pts = {{0, 0}, {1, 0.5`}, {1, 1}, {0.7`, 0.5`}}
*)
(*
SimplePolygonTriangulation[pts, GraphicalDebugging -> True]
*)
(*
gra = Graphics[Polygon[pts]];
*)
(*
(* simple example with SimplePolygonTriangulate : *)
*)
(*
Show[TriangulateSimplePolygons[gra, GraphicalDebugging -> True]];
*)
(*
pts = {{0, 0}, {1, 0}, {0, 1}, {1, 1}}
*)
(*
(* error with winding number : *)
*)
(*
SimplePolygonTriangulation[pts]
*)
(*
(* error with self intersection : *)
*)
(*
pts = {{0, 0}, {1, 0}, {0, 1}, {0.5, -0.5}, {0, 1}}
*)
(*
SimplePolygonTriangulation[pts]
*)
(*
pts = {{0, 0, 0}, {1, 0, 0}, {1, 1, 0}, {0.7, 0.5, 0}}
*)
(*
SimplePolygonTriangulation3D[pts, GraphicalDebugging -> True, 
  PlaneNormal -> {-1, 1, 1}]
*)
(*
TriangulateSimplePolygons3D[Graphics3D[Polygon[pts]], 
  GraphicalDebugging -> True, QuadrangleQuartering -> False]
*)
(*
(* spirals (worst case examples) : *)
*)
(*
Spiral[n_, m_] := 
  Module[{pts}, 
    pts = Table[{phi{Cos[phi], Sin[phi]}, (phi + Pi){Cos[phi], 
              Sin[phi]}}, {phi, 0, n 2Pi, 2. Pi/m}]; 
    pts = RotateLeft[
        Flatten[{Transpose[pts][[1]], Reverse[Transpose[pts][[2]]]}, 1], 
        Floor[m*n/2]]]
*)
(*
times = {};
Do[pts = Spiral[n, 6]; 
    times = Append[
        times, {Length[pts], 
          Timing[SimplePolygonTriangulation[pts, 
                  GraphicalDebugging -> False]][[1]] /. Second -> 1}]; 
    Print[times], {n, 1, 100, 10}];
*)
(*
fitted = Fit[times, {x, x Log[x]}, x]
*)
(*
Show[Plot[fitted, {x, 0, Last[times][[1]]}], ListPlot[times]]
*)
(*
(* fractal test (another worst - case) : *)
*)
(*
iterate[a_, b_, c_, d_, level_, maxlevel_] := 
  Module[{ab1 = a + 0.45(b - a), ab2 = a + 0.55(b - a), 
      leftDir = 0.5*{{0, -1}, {1, 0}}.(b - a), cd1 = c + 0.45(d - c), 
      cd2 = c + 0.55(d - c), rightDir = 0.5*{{0, -1}, {1, 0}}.(d - c)}, 
    If[level < maxlevel, 
      Flatten[{{a}, 
          iterate[ab1, ab1 + leftDir, ab2 + leftDir, ab2, level + 1, 
            maxlevel], {b, c}, 
          iterate[cd1, cd1 + rightDir, cd2 + rightDir, cd2, level + 1, 
            maxlevel], {d}}, 1], {a, b, c, d}]]
*)
(*
fractalCross[n_] := iterate[{0, 0}, {0, 1}, {0.1, 1}, {0.1, 0}, 0, n]
*)
(*
times = {};
Do[pts = fractalCross[n]; 
    times = Append[
        times, {Length[pts], 
          Timing[SimplePolygonTriangulation[pts]][[1]] /. Second -> 1}]; 
    Print[times], {n, 3, 8}];
*)
(*
fitted = Fit[times, {1, x, x Log[x]}, x]
*)
(*
Show[Plot[fitted, {x, 0, Last[times][[1]]}], ListPlot[times]];
*)
(*
(* random test : *)
*)
(*
size = 10;
*)
(*
pts = Flatten[{Table[{{i, 0}, {i + 0.5, size*Random[]}}, {i, 0, size}], 
      Table[{{i + 0.5, size}, {i, size*Random[]}}, {i, size, 0, -1}]}, 2]
*)
(*
SimplePolygonTriangulation[pts, GraphicalDebugging -> True];
*)
(*
(* Plot3D test *)
*)
(*
gr = Graphics3D[Plot3D[Sin[x]Cos[y], {x, 0, 10}, {y, 0, 10}]];
*)
(*
Show[TriangulateSimplePolygons3D[gr, QuadrangleQuartering -> True, 
      TriangleEdges -> Automatic]];
*)
(*
(* non - convex test *)
*)
(*
g = Show[Graphics3D[{Polygon[
            Table[(2 + 0.5Sin[3 phi] + 0.5Sin[10  phi]){Cos[phi], Sin[phi], 
                  0}, {phi, 0., 2. Pi, 2. Pi/100}]]}]];
*)
(*
Show[TriangulateSimplePolygons3D[g, GraphicalDebugging -> False]];
*)
(*
Show[TriangulateSimplePolygons3D[g, TriangleEdges -> All]];
*)
(*
(* non - flat, non - convex test *)
*)
(*
g = Show[Graphics3D[{FaceForm[SurfaceColor[RGBColor[0, 0, 1]], 
            SurfaceColor[RGBColor[1, 1, 0]]], 
          Polygon[Table[(2 + 0.5Sin[3 phi] + 0.5Sin[10  phi]){Cos[phi], 
                  Sin[phi], Cos[phi]^2}, {phi, 0., 2. Pi, 2. Pi/100}]]}]];
*)
(*
Show[TriangulateSimplePolygons3D[g, TriangleEdges -> All]];
*)
(*
Show[TriangulateSimplePolygons3D[g, TriangleEdges -> All], 
    ViewPoint -> {1, 1, 0}];
*)
